#include "BinaryHash.h"

BinaryHash::BinaryHash(int _dim, METHOD_ID _method_id) 
	: dim(_dim), method_id(_method_id) 
{
	p_matrix = new REAL_TYPE*[dim];

	if (method_id == SBQ) 
	{
		p_num = BCODE_LEN;
		for (int i = 0; i < dim; i++) 
			p_matrix[i] = new REAL_TYPE[p_num];
	} 
	else
	{
		p_num = BCODE_LEN / 2;
		for (int i = 0; i < dim; i++)
			p_matrix[i] = new REAL_TYPE[p_num];

		threshold1 = new REAL_TYPE[p_num];
		threshold2 = new REAL_TYPE[p_num];
		threshold3 = new REAL_TYPE[p_num];
	}
}

BinaryHash::~BinaryHash() 
{
	for (int i = 0; i < dim; i++)
		delete[] p_matrix[i];

	delete[] p_matrix;

	if (method_id != SBQ) 
	{
		delete[] threshold1;
		delete[] threshold2;
		delete[] threshold3;
	}
}

void BinaryHash::initialize_LSH()
{
	for (int i = 0; i < dim; i++)
	{
		for (int j = 0; j < p_num; j++)
		{
			p_matrix[i][j] = Rand_Gaussian<REAL_TYPE>();
		}
	}
}

void BinaryHash::initialize_matrix(char* pM_path) {
	FILE *fin;
	fin = fopen(pM_path, "r");
	for (int i = 0; i < dim; i++)
	{
		for (int j = 0; j < p_num; j++)
		{
			fscanf(fin, "%f", &p_matrix[i][j]);
		}
	}
	fclose(fin);
}

void BinaryHash::compute_BCode(REAL_TYPE *x, bitset<BCODE_LEN> &y)
{
	if (method_id == SBQ) 
	{
		REAL_TYPE tmp;
		for (int i = 0; i < p_num; i++)
		{
			tmp = 0.0;
			for (int k = 0; k < dim; k++)
			{
				tmp += x[k] * p_matrix[k][i];
			}
		
			if (tmp > 0.0)
			{
				y[i] = 1;
			}
			else
			{
				y[i] = 0;
			}
		}
	}
	else if (method_id == DBQ)
	{
		REAL_TYPE tmp;
		for (int i = 0; i < p_num; i++)
		{
			tmp = 0.0;
			for(int k = 0; k < dim; k++)
			{
				tmp += x[k] * p_matrix[k][i];
			}

			if(tmp <= threshold1[i])
			{
				y[i] = 0;
				y[i + p_num] = 1;
			}
			else if (tmp > threshold3[i])
			{
				y[i] = 1;
				y[i + p_num] = 0;
			} 
			else
			{
				y[i] = 0;
				y[i + p_num] = 0;
			}
		}
	}
	else if (method_id == QE)
	{
		REAL_TYPE tmp;
		for (int i = 0; i < p_num; i++)
		{
			tmp = 0.0;
			for(int k = 0; k < dim; k++)
			{
				tmp += x[k] * p_matrix[k][i];
			}

			if(tmp > threshold2[i])
			{
				y[i] = 1;
			}
			else
			{
				y[i] = 0;
			}

			if(threshold1[i] < tmp && tmp < threshold3[i])
			{
				y[i + p_num] = 0;
			}
			else
			{
				y[i + p_num] = 1;
			}
		}
	}
}


// training Quadra-Embedding thresholds
// takes O(n^2)
void BinaryHash::train_QE(int num_train_samples, Points &tps)
{
	REAL_TYPE *projection_values;
	projection_values = new REAL_TYPE[num_train_samples];		

	REAL_TYPE **var;
	var = new REAL_TYPE*[num_train_samples];

	for (int i = 0; i < num_train_samples; i++) 
	{
		var[i] = new REAL_TYPE[num_train_samples];
	}

	// variables for parallelization
	double *min_values;
	REAL_TYPE *min_as, *min_bs, *min_cs;
	int *as, *bs, *cs;

	min_values = new double[num_train_samples];
	min_as = new REAL_TYPE[num_train_samples];
	min_bs = new REAL_TYPE[num_train_samples];
	min_cs = new REAL_TYPE[num_train_samples];
	as = new int[num_train_samples];
	bs = new int[num_train_samples];
	cs = new int[num_train_samples];


	// compute thresholds
	for (int i = 0; i < p_num; i++) 
	{
		#ifdef USE_PARALLELIZATION		
		#pragma omp parallel for
		#endif
		for (int j = 0; j < num_train_samples; j++) 
		{
			projection_values[j] = 0;
			for(int k = 0; k < dim; k++)
			{
				projection_values[j] += tps.d[j][k] * p_matrix[k][i];
			}
		}
		sort(projection_values, projection_values + num_train_samples);

		#ifdef USE_PARALLELIZATION		
		#pragma omp parallel for
		#endif
		for (int j = 0; j < num_train_samples; j++) 
		{
			var[j][j] = 0;

			REAL_TYPE sum_remain = projection_values[j];
			REAL_TYPE sum_sqr = sqr(projection_values[j]);
			REAL_TYPE sum_all = projection_values[j];
			int cnt_remain = 1;
			int cnt_all = 1;
			int idx_m = j;

			// Using mean
			for (int k = j + 1; k < num_train_samples; k++) 
			{
				REAL_TYPE new_mean = (sum_all + projection_values[k]) / (REAL_TYPE)(cnt_all + 1);
				sum_all += projection_values[k];
				sum_remain += projection_values[k];
				sum_sqr += sqr(projection_values[k]);
				cnt_remain ++;
				while (projection_values[idx_m] < new_mean) 
				{
					sum_sqr -= sqr(projection_values[idx_m]);
					sum_remain -= projection_values[idx_m];					
					cnt_remain --;
					idx_m ++;
				}
				var[j][k] = sum_sqr - 2 * new_mean * sum_remain + sqr(new_mean) * cnt_remain;
				cnt_all ++;
			}

			sum_remain = projection_values[j];
			sum_sqr = sqr(projection_values[j]);
			sum_all = projection_values[j];
			cnt_remain = 1, cnt_all = 1, idx_m = j;

			for (int k = j - 1; k >= 0; k--) 
			{
				REAL_TYPE new_mean = (sum_all + projection_values[k]) / (REAL_TYPE)(cnt_all + 1);
				sum_all += projection_values[k];
				sum_remain += projection_values[k];
				sum_sqr += sqr(projection_values[k]);
				cnt_remain ++;
				while (new_mean < projection_values[idx_m]) 
				{
					sum_sqr -= sqr(projection_values[idx_m]);
					sum_remain -= projection_values[idx_m];					
					cnt_remain --;
					idx_m --;
				}
				var[j][k] = sum_sqr - 2 * new_mean * sum_remain + sqr(new_mean) * cnt_remain;
				cnt_all ++;
			}
		}

		double min_value = 100000000.0;
		double min_a, min_b, min_c;
		int a, b, c;

		for (int j = 0; j < num_train_samples; j++)
			min_values[j] = 100000000.0;

		#ifdef USE_PARALLELIZATION		
		#pragma omp parallel for
		#endif
		for (int t1 = 10; t1 < num_train_samples - 10; t1++) 
		{
			int t2 = t1 + 100;
			for (int t3 = t1 + 100; t3 < num_train_samples - 10; t3++) 
			{
				double tmp = 0;
				tmp += var[0][t1];
				tmp += var[num_train_samples - 1][t3];

				/*t2 = (t1 + t3) / 2;*/
				double min_var = var[t2][t1] + var[t2 + 1][t3];
				while (t2 + 50 < t3 && min_var > var[t2 + 1][t1] + var[t2 + 2][t3]) 
				{
					t2 ++;
					min_var = var[t2][t1] + var[t2 + 1][t3];
				}


				tmp += min_var;

				if (min_values[t1] > tmp) 
				{
					min_values[t1] = tmp;
					min_as[t1] = projection_values[t1];
					min_bs[t1] = projection_values[t2];
					min_cs[t1] = projection_values[t3];
					as[t1] = t1;
					bs[t1] = t2;
					cs[t1] = t3;
				}
			}
		}

		for (int t1 = 0; t1 < num_train_samples; t1++) 
		{
			if (min_value > min_values[t1]) 
			{
				min_value = min_values[t1];
				min_a = min_as[t1];
				min_b = min_bs[t1];
				min_c = min_cs[t1];
				a = as[t1];
				b = bs[t1];
				c = cs[t1];
			}
		}

		threshold1[i] = min_a;
		threshold2[i] = min_b;
		threshold3[i] = min_c;
//			printf("\t\tThresholds (%d): %.4lf (%d)\t%.4lf (%d)\t%.4lf (%d)\n", i, threshold1[i], a, threshold2[i], b, threshold3[i], c);
	}

	for (int i = 0; i < num_train_samples; i++) 
	{
		delete [] var[i];
	}
	delete [] var;
	delete [] min_values;
	delete [] min_as;
	delete [] min_bs;
	delete [] min_cs;
	delete [] as;
	delete [] bs;
	delete [] cs;
	delete [] projection_values;
}

// training DBQ thresholds
// takes O(n log n)
void BinaryHash::train_DBQ(int num_train_samples, Points &tps)
{
	REAL_TYPE *projection_values;
	projection_values = new REAL_TYPE[num_train_samples];		
		
	// compute thresholds
	for (int i = 0; i < p_num; i++) 
	{
		#ifdef USE_PARALLELIZATION		
		#pragma omp parallel for
		#endif
		for (int j = 0; j < num_train_samples; j++) 
		{
			projection_values[j] = 0;
			for(int k = 0; k < dim; k++)
			{
				projection_values[j] += tps.d[j][k] * p_matrix[k][i];
			}
		}
		sort(projection_values, projection_values + num_train_samples);

		double sum1, sum2, sum3, F_max;
		int p1, p2, a, b;
		p1 = p2 = 0;
		sum1 = sum2 = sum3 = 0.0;
		F_max = 0.0;

		for (int j = 0; j < num_train_samples; j++) 
		{
			sum1 += projection_values[j];
			if (projection_values[j] > 0) 
			{
				p1 = j - 1;
				break;
			}
		}
		p2 = p1 + 1;
		for (int j = p2; j < num_train_samples; j++)
			sum3 += projection_values[j];

		while (p1 != -1 && p2 != num_train_samples) 
		{
			if (sum2 <= 0) 
			{
				sum2 += projection_values[p2];
				sum3 -= projection_values[p2];
				p2 ++;
			} else 
			{
				sum2 += projection_values[p1];
				sum1 -= projection_values[p1];
				p1 --;
			}
			if (p1 != -1 && p2 != num_train_samples) 
			{
				double F = sum1 * sum1 / (p1 + 1) + sum3 * sum3 / (num_train_samples - p2);
				if (F > F_max) 
				{
					F_max = F;
					a = p1;
					b = p2 - 1;
				}
			}
		}

		threshold1[i] = projection_values[a];
		threshold2[i] = 0.0;
		threshold3[i] = projection_values[b];
	}

	delete [] projection_values;
}
