///////////////////////////////////////////////////
//  Sample code for 'Quadra-Embedding: Binary    //
//  Code Embedding with Low Quantization Error'  //
//  of ACCV 2012 paper.                          //
//  http://sglab.kaist.ac.kr/quadra			     //
//  Youngwoon Lee, 2012-10-19					 //
///////////////////////////////////////////////////

///////////////////////////////////////////////////
//  Built on Jae-Pil's Spherical Hashing code    //
//  http://sglab.kaist.ac.kr/SphericalHashing    //
///////////////////////////////////////////////////

#include "BinaryHash.h"
#include "Evaluation.h"

#ifdef USE_PARALLELIZATION
#include <omp.h>
#endif

Points dps;				// dps: data points set
Points qps;				// qps: query points set
Points tps;				// tps: training points set
int nP;					// nP: number of data points
int nQ;					// nQ: number of query points

int **gt;				// gt: ground truth labels

REAL_TYPE *data_center;	// center of data points


// initialize data and query points
void initialize_data()
{
	dps.initialize_from_file(INPUT_DATA_FILE_NAME);
	qps.initialize_from_file(INPUT_QUERY_FILE_NAME);
	dps.sample_points(NUM_TRAIN_SAMPLES, &tps);
	
	nP = dps.nP;
	nQ = qps.nP;

	// you can control the number of queries here
#ifdef NUM_QUERIES
	nQ = NUM_QUERIES;
#endif

	// compute mean position of data points
	data_center = new REAL_TYPE[dps.dim];
	dps.compute_center(data_center);
}


// zero-centering
void do_zero_centering()
{
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for (int i = 0; i < nP; i++)
	{
		Sub_Vector<REAL_TYPE>(dps.d[i], data_center, dps.d[i], dps.dim);
	}
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for (int i = 0; i < nQ; i++)
	{
		Sub_Vector<REAL_TYPE>(qps.d[i], data_center, qps.d[i], qps.dim);
	}
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for (int i = 0; i < NUM_TRAIN_SAMPLES; i++)
	{
		Sub_Vector<REAL_TYPE>(tps.d[i], data_center, tps.d[i], tps.dim);
	}
}

// undo zero-centering
void undo_zero_centering()
{
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for (int i = 0; i < nP; i++)
	{
		Add_Vector<REAL_TYPE>(dps.d[i], data_center, dps.d[i], dps.dim);
	}
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for (int i = 0; i < nQ; i++)
	{
		Add_Vector<REAL_TYPE>(qps.d[i], data_center, qps.d[i], qps.dim);
	}
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for (int i = 0; i < NUM_TRAIN_SAMPLES; i++)
	{
		Add_Vector<REAL_TYPE>(tps.d[i], data_center, tps.d[i], tps.dim);
	}
}

// compute ground-truth (time comsuming)
// I recommend you to store ground truth information to the file 
void compute_ground_truth()
{
	gt = new int *[nQ];
	for (int i = 0; i < nQ; i++)
	{
		gt[i] = new int[K];
	}
	Result_Element<REAL_TYPE> *tmp = new Result_Element<REAL_TYPE>[nP];
	for (int i = 0; i < nQ; i++)
	{
#ifdef USE_PARALLELIZATION
		#pragma omp parallel for
#endif
		for (int j = 0; j < nP; j++)
		{
			tmp[j].index = j;
			tmp[j].dist = Compute_Distance_L2Sq<REAL_TYPE>(qps.d[i], dps.d[j], dps.dim);
		}
		sort(&tmp[0], &tmp[nP]);
#ifdef USE_PARALLELIZATION
		#pragma omp parallel for
#endif
		for (int j = 0; j < K; j++)
		{
			gt[i][j] = tmp[j].index;
		}
	}
	delete[] tmp;
}

#ifdef INPUT_GROUND_TRUTH_FILE_NAME
void load_ground_truth()
{
	FILE *input;
	input = fopen(INPUT_GROUND_TRUTH_FILE_NAME, "rb");
	
	int _nQ;
	int _K;
	fread(&_nQ, sizeof(int), 1, input);
	fread(&_K, sizeof(int), 1, input);

	gt = new int*[_nQ];
	for (int i = 0; i < _nQ; i++)
	{
		gt[i] = new int[_K];
	}

	for (int i = 0; i < _nQ; i++) {
		for (int j = 0; j < _K; j++) {
			float tmp;
			fread(&gt[i][j], sizeof(int), 1, input);
			fread(&tmp, sizeof(float), 1, input); // dummy for distance
		}
	}
	fclose(input);
}
#endif

void evaluate(const char *method_name, BinaryHash &hash) {
	printf("%s :\n", method_name);

	Stopwatch T0("");
	bitset<BCODE_LEN> *bCodeData = new bitset<BCODE_LEN> [nP];
	bitset<BCODE_LEN> *bCodeQuery = new bitset<BCODE_LEN> [nQ];
	Result_Element<int> *res = new Result_Element<int> [nP];


	// training hashing functions
	if (hash.method_id == QE)
	{
		printf("\t- Training QE can take several minutes\n");
	}

	T0.Reset();
	T0.Start();
	printf("\t- Training Hash Functions ");
	if (hash.method_id == DBQ)
	{
		hash.train_DBQ(NUM_TRAIN_SAMPLES, tps);
	}
	else if (hash.method_id == QE)
	{
		hash.train_QE(NUM_TRAIN_SAMPLES, tps);
	}
	T0.Stop();
	printf("Finished (%f seconds)\n", T0.GetTime());


	// compute binary codes
	printf("\t- Computing Binary Codes ");
	T0.Reset();		
	T0.Start();
	
#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for(int i=0;i<nP;i++)
	{
		hash.compute_BCode(dps.d[i], bCodeData[i]);
	}

#ifdef USE_PARALLELIZATION
	#pragma omp parallel for
#endif
	for(int i=0;i<nQ;i++)
	{
		hash.compute_BCode(qps.d[i], bCodeQuery[i]);
	}
	T0.Stop();
	printf("Finished (%f seconds)\n", T0.GetTime());


	// process queries
	printf("\t- Evaluating results ");
	T0.Reset();
	T0.Start();
	double mAP = 0.0;	
	for (int qIndex = 0; qIndex < nQ; qIndex++)
	{
		if (hash.method_id == SBQ || hash.method_id == DBQ) 
		{
#ifdef USE_PARALLELIZATION		
			#pragma omp parallel for
#endif
			for (int i = 0; i < nP; i++)
			{
				res[i].index = i;
				res[i].dist = compute_HD(bCodeQuery[qIndex], bCodeData[i]);
			}
		}
		else
		{
#ifdef USE_PARALLELIZATION		
			#pragma omp parallel for
#endif
			for (int i = 0; i < nP; i++)
			{
				res[i].index = i;
				res[i].dist = compute_QED(bCodeQuery[qIndex], bCodeData[i]);
			}
		}
		mAP += Compute_AP<int>(gt[qIndex], res, nP);
	}
	mAP /= (double)nQ;

	printf("Finished (%f seconds)\n", T0.GetTime());
	printf("\t- mAP : %f\n", mAP);
	printf("\n");

	delete[] bCodeData;
	delete[] bCodeQuery;
	delete[] res;
}

void process()
{
	BinaryHash LSH(dps.dim, SBQ);
	BinaryHash LSH_DBQ(dps.dim, DBQ);
	BinaryHash LSH_QE(dps.dim, QE);

	LSH.initialize_LSH();
	LSH_DBQ.initialize_LSH();
	LSH_QE.initialize_LSH();

#ifdef INPUT_SBQ_MATRIX_FILE_NAME
	BinaryHash ITQ(dps.dim, SBQ);
	ITQ.initialize_matrix(INPUT_SBQ_MATRIX_FILE_NAME);
#endif
#ifdef INPUT_DBQ_MATRIX_FILE_NAME
	BinaryHash ITQ_DBQ(dps.dim, DBQ);
	BinaryHash ITQ_QE(dps.dim, QE);
	ITQ_DBQ.initialize_matrix(INPUT_DBQ_MATRIX_FILE_NAME);
	ITQ_QE.initialize_matrix(INPUT_DBQ_MATRIX_FILE_NAME);
#endif


	do_zero_centering();

	printf("\n");
	printf("- Results\n");
	printf("\n");

	evaluate("LSH (SBQ)", LSH);
	evaluate("LSH (DBQ)", LSH_DBQ);
	evaluate("LSH (QE)", LSH_QE);

#ifdef INPUT_SBQ_MATRIX_FILE_NAME
	evaluate("ITQ (SBQ)", ITQ);
#endif
#ifdef INPUT_DBQ_MATRIX_FILE_NAME
	evaluate("ITQ (DBQ)", ITQ_DBQ);
	evaluate("ITQ (QE)", ITQ_QE);
#endif

	undo_zero_centering();
}

int main()
{	
	srand((unsigned int)time(NULL));
	printf("\n");
	printf("---   Simple Version of Quadra-Embedding Simulation   ---\n");
	printf("\t- Only support K-NN experiment with K = %d\n", K);
	printf("\t- Code length: %d bits\n", BCODE_LEN);
	printf("\t- # training samples: %d\n", NUM_TRAIN_SAMPLES);
	printf("\n");


	// read data
	printf("- Reading Data ");
	Stopwatch T0("");
	T0.Reset();
	T0.Start();
	
	initialize_data();
	
	T0.Stop();
	printf("Finished (%f seconds)\n", T0.GetTime());

	
	// compute ground truth
	printf("- Computing GroundTruth ");
	T0.Reset();
	T0.Start();
	
#ifdef INPUT_GROUND_TRUTH_FILE_NAME
	load_ground_truth();
#else
	compute_ground_truth();
#endif

	T0.Stop();
	printf("Finished (%f seconds)\n", T0.GetTime());


	// evaluate
	T0.Reset();
	T0.Start();
	
	process();
	
	T0.Stop();
	printf("- Evaluating All Methods Finished (%f seconds)\n", T0.GetTime());


	dps.release_memory();
	qps.release_memory();
	tps.release_memory();
	for (int i = 0; i < nQ; i++)
	{
		delete[] gt[i];
	}
	delete[] gt;
	delete[] data_center;
	
	return 0;
}